#!/usr/bin/env python
"""

Command line utility to help build maps with the Generic Mapping Tool
(GMT) package for file and trace header information stored in an
SQLite database.

     Copyright (C) 2005  Kurt Schwehr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

$Id: segysqlgmt.py,v 1.19 2005/09/15 16:06:05 schwehr Exp $

http://www.python.org/doc/lib/module-optparse.html

http://schwehr.org/software/segy-py/
"""

# These values end up in the pydoc web page
__version__ = '$Revision: 1.19 $' # See man ident
__date__ = '$Date: 2005/09/15 16:06:05 $' # FIX: pull out just the date
__author__ = 'Kurt Schwehr'
__credits__ = """Lots of people for helping me, especially Neal
Driscoll and Lisa Tauxe."""

import sqlite
import os
import sys

BOMBASTIC= 4
VERBOSE  = 3
TRACE    = 2
TERSE    = 1
ALWAYS   = 0


from optparse import OptionParser

vcvs = "$Revision: 1.19 $"
vcvs = vcvs[vcvs.find(':')+2:]
vcvs = vcvs[:vcvs.find('$')-1]
version_cvs = vcvs
del (vcvs) # stop polution

# Track the version of segy-py.  If I use setup.py, which one gets imported?
from segy_version import VERSION

myparser = OptionParser(usage="%prog [options]",
                      version="%prog "+VERSION+' - '+version_cvs)
myparser.add_option('-d','--database',dest='dbfilename',default='segy.db',
                    help='Database containing imported segy file and trace header info [default: %default]')
myparser.add_option('-m','--min-file-key',dest='minFileKey',default=None, type='int',
                    help='Set this to start processing at a certain file index [default: %default]')
myparser.add_option('-M','--max-file-key',dest='maxFileKey',default=None, type='int',
                    help='Set this to end processing at a certain file index [default: %default]')


myparser.add_option('-F','--fledermaus',dest='fledermaus',default=None, type='str',
                    help='Output a text file suitable for IVS Fledermaus.  Skips other types. [default: %default]')
myparser.add_option('-P','--fledermaus-pts',dest='fledermausPts',default=None, type='str',
                    help='Output a text file suitable for IVS Fledermaus point import.  Uses --key.  Skips other types. [default: %default]')


# FIX: Option to skip
myparser.add_option('-l','--line-file',dest='shotLineFilename',default='shot-locations.psxy',
                    help='Where to write the unique shot location line info [default: %default]')
myparser.add_option('-L','--skip-line',dest='skipLine',default=False,action="store_true",
                    help='Skip the shot line points [default: %default]')

# No longer exists...
#myparser.add_option('-b','--batch-lines',dest='batchLineFiles',default=False,action='store_true',
#                    help='Process all the files for lines together without picking up the pen [default: %default]')

# FIX: Option to force hemisphere to cope with xstar missing longitude missing negatives

# FIX: Option to skip
myparser.add_option('-s','--shot-label-file',dest='shotTextFilename',default='shot-labels.pstext',
                    help='Where to write the shot labels for every Nth shot [default: %default]')
myparser.add_option('-n','--nth-shot-label',dest='nthShotLabel',type='int',default='2000',
                    help='How many shots to skip before writing out a label [default: %default]')
myparser.add_option('-a','--label-angle',dest='shotLabelAngle',default='45',
                    help='How much to angle the text for each shot label [default: %default]')
myparser.add_option('-f','--font-size',dest='fontSize',default='1',
                    help='Size of the label font in points [default: %default]')

myparser.add_option('-g','--gmt-script',dest='makeGmtScript',default=False,action='store_true',
                    help='Print out example GMT script code')

myparser.add_option('-S','--scaler-coord',dest='scalerCoord',default=False,action='store_true',
                    help='Divide by the scaler coord field in the trace header to get the true lat/lon')

myparser.add_option('-x','--flip-x',dest='flipX',default=False,action='store_true',
                    help='Multiply the x coordinate by -1')
myparser.add_option('-y','--flip-y',dest='flipY',default=False,action='store_true',
                    help='Multiply the x coordinate by -1')

myparser.add_option('-k','--key',dest='key',default='Shotpoint', type='string',
                    help='segyTrace label to use for the shot text.  e.g. traceNumber or EnsembleNo [default: %default]')


myparser.add_option('-v',action="count",dest='verbosity',default=0,
                    help='how much information to give.  Specify multiple times to increase verbosity')
myparser.add_option('--verbosity',dest='verbosity',type='int',
                    help='Specify verbosity.  Should be in the range of '
                    +str(ALWAYS)+'...'+str(BOMBASTIC)+' (None...Bombastic)')

myparser.add_option('-c','--coord-unit',type="int",dest='forceCoordUnit',default=-1,
                    help='Force a SEGY coordinate system, e.g. 2 == seconds of arc.  Use this for short format databases. [default: follow file]')

myparser.add_option('-H','--force-hemisphere',type="int",dest='forceHemisphere',default=-1,
                    help='Force hemisphere for shots that have bad navigation.'+
                    '  0-NW,  1-NE,  2-SW,  3-SE')

myparser.add_option('-u','--from-utm',action="store_true",dest='fromUTM',default=False,
                    help='Assume x and y are in UTM and convert to Long/Lat')
myparser.add_option('-z','--utm-zone',type="int",dest='utmZone',default=None,
                    help='Specify the UTM zone if converting from UTM to Long/Lat')


myparser.add_option('-Z','--skip-zeros',default=False,dest='skipZeros',action='store_true',
                    help='Skip points that have 0 0 stored in the x,y fields')




def forceHemisphere(x,y,hemisphere):
    """
    x - longtidude (float)
    y - latitude (float)
    hemisphere: -1 no change
                0 NW
                1 NE
                2 SW
                3 SE
    ASSUMES: lat/lon with lon in [-180..180]
    """
    #print "hemi: ",x,y,hemisphere
    if -1==hemisphere: return x,y
    x = float(x)
    y = float(y)
    if 0==hemisphere: # NW
        if (x>0): x = -x
        if (y<0): y = -y
    elif 1==hemisphere: # NE
        if (x<0): x = -x
        if (y<0): y = -y
    elif 2==hemisphere: # SW
        if (x>0): x = -x
        if (y>0): y = -y
    elif 3==hemisphere: # SE
        if (x>0): x = -x
        if (y<0): y = -y
    else:
        assert(False and "ERROR: invalid hemisphere number")
    return x,y

# FIX: allow the code to detect location X,Y format.  Do not force /3600.0

def buildShotLines(db_cx, options, outFilename='monkey-boy.pstext'):
    """
    Emit series of points so that we can plot up the lines in GMT
    """

    if options.verbosity>=VERBOSE: print 'file key ranges = ',options.minFileKey,options.maxFileKey,
    cu = db_cx.cursor()
    ct = CoordTransform(cx,options)
    out = open(outFilename,'w')

       
    searchStr  = "SELECT fileKey,x,y FROM segyTrace"
    if (None != options.minFileKey) or (None != options.maxFileKey): searchStr += ' WHERE'
    if (None != options.minFileKey): searchStr += ' fileKey>='+str(options.minFileKey)
    if (None != options.minFileKey) and (None != options.maxFileKey): searchStr += ' and'
    if (None != options.maxFileKey): searchStr += ' fileKey<='+str(options.maxFileKey)
    searchStr += ';'

    if options.verbosity >= VERBOSE: print searchStr 
    cu.execute(searchStr)

    lastx = lasty = -1
    warnedAboutRange = False
    lastFileKey = -1
    for row in cu.fetchall():
        x = row['x']
        y = row['y']
        if 0 == x and 0 == y:
            if options.verbosity>=TRACE: print "Skipping possible bad value"
            continue

        # Make sure to separate files
        fileKey = row['fileKey']
        if options.verbosity>=BOMBASTIC*2: print '  file keys:  ',lastFileKey, fileKey
        if fileKey != lastFileKey:
            if -1 != lastFileKey:
                out.write('>\n')
                if options.verbosity>=TRACE: print 'marking file separation',lastFileKey, fileKey
            lastFileKey = fileKey
            fileOffset = 0

        x,y = ct.getLonLat(x,y)
        
        if 180<y or -180>y or -360>x or 360<x:
            if not warnedAboutRange:
                print "FIX: range error", x,y
                warnedAboutRange = True  # only warn once

        if lastx != x or lasty != y:
            # Only write out a line when we get a new location
            out.write(str(x)+" "+str(y)+"\n")
            lastx = x
            lasty = y

    out.close()
    return


def buildShotText(db_cx, options, outFilename='JohnYahYah.pstext'):
    """
    Produce pstext input file for labeling every nth shot.

    """
    if options.verbosity >= TERSE:
        print "buildShotText",outFilename,options.nthShotLabel,options.shotLabelAngle

    cu = db_cx.cursor()
    ct = CoordTransform(cx,options)
    if options.verbosity >= TRACE: print "outFilename: ",outFilename
    out = open(outFilename,'w')

    searchStr = "SELECT fileKey,"+options.key+" as key,x,y FROM segyTrace "

    searchStr += "WHERE key % "+str(options.nthShotLabel)+"= 1" # FIX: need  ORDER BY shotpoint???. 

    if (None != options.minFileKey): searchStr += ' fileKey>='+str(options.minFileKey)
    if (None != options.minFileKey) and (None != options.maxFileKey): searchStr += ' and'
    if (None != options.maxFileKey): searchStr += ' fileKey<='+str(options.maxFileKey)
    searchStr += ';'

    if options.verbosity >= VERBOSE:  print "Seach string: ",searchStr 
    cu.execute(searchStr)
    
    angle = str(options.shotLabelAngle)
    found = False
    for row in cu.fetchall():
        found = True

        x = row['x']
        y = row['y']
        if 0 == x and 0 == y: continue # skip 0,0 lines

        shotpoint = str(row['key'])
        fileKey   = str(row['fileKey'])

        x,y = ct.getLonLat(x,y)
        out.write(str(x)+" "+str(y)+' '+str(options.fontSize)+' '+str(angle)+' 1 LM '+str(fileKey)+'-'+str(shotpoint)+'\n')

    out.close()
    if not found and options.verbosity >= TRACE:
        print '\nWARNING: No shot labels were generated.\n  Suggest adding "--key=traceNumber"\n'
    return

##################################################

class CoordTransform:
    def __init__(self, cx, options):
        if options.verbosity>=TRACE:
            print 'init CoordTransform'

        cu = cx.cursor()  # Database cursor

        if -1 == options.forceCoordUnit:
            cu.execute('SELECT CoordUnits FROM segyTrace where traceNumber=1')
            self.coordUnits = int(cu.fetchone()['CoordUnits'])
        else:
            self.coordUnits = options.forceCoordUnit
        if options.verbosity >= VERBOSE:
            coordUnitsTbl = {
                1: 'Length (meters or feet)',
                2: 'Seconds of arc',
                3: 'Decimal degrees',
                4: 'Degrees, minutes, seconds'
                }
            print 'coordUnits = ',self.coordUnits, ' -- ',
            try:
                print coordUnitsTbl[self.coordUnits]
            except KeyError:
                print 'Unknown coord units... good luck bub'

        self.scaleUnits = 1
        if self.coordUnits == 2: self.scaleUnits = 1/3600.  # 2 is Seconds of arc
        if self.coordUnits == 4: print "WARNING: degree, min, sec not handled"
        # Otherwise NOP

        # FIX: if ScalerCoord varies between files, we are screwed!
        cu.execute('SELECT ScalerCoord FROM segyTrace WHERE traceNumber=1')
        self.scalerCoord = float(cu.fetchone()['ScalerCoord'])
        if self.scalerCoord == 0.:
            self.scalerCoord = 1
            if options.verbosity>=TERSE: print 'Forcing scalerCoord to be 1.  Zero is NOT OK.'
        if options.verbosity>=TRACE: print 'ScalerCoord = ',self.scalerCoord

        self.fromUTM = options.fromUTM
        if self.fromUTM:
            if None == options.utmZone:
                sys.exit("ERROR: you must specify the UTM zone")
            if (type(10) != type(options.utmZone)):
                sys.exit("FIX: zone must be an int")
            from pyproj import Proj
            params = {}
            params['proj'] = 'utm'
            params['zone'] = options.utmZone
            self.proj = Proj(params)

        self.flipX = options.flipX
        self.flipY = options.flipY

        self.forceHemisphere = options.forceHemisphere

    def getLonLat(self, x,y):
        """Use the info collected by init to convert x,y to longitude,latitude floats"""
        x = float(x) * self.scaleUnits / self.scalerCoord
        y = float(y) * self.scaleUnits / self.scalerCoord

        if self.fromUTM: x,y = self.proj(x,y,inverse=True)

        if self.flipX: x = -x
        if self.flipY: y = -y

        if self.forceHemisphere != -1: x,y = forceHemisphere(x,y,self.forceHemisphere)

        lon = x
        lat = y
        return lon,lat

##################################################

def buildFledermausLines(cx, options):
    """
    Build a fledermaus x,y file with 9999,9999 sentinal values

    cx - database connection
    options.fledermaus - file to write out to
    options.nthShotLabel - how often to write out points
    """

    out = open(options.fledermaus,'w')

    ct = CoordTransform(cx,options)
    cu = cx.cursor()  # Database cursor

    if options.verbosity >= TRACE:
        print 'fileKeys: ', options.minFileKey, options.maxFileKey
    
    searchStr  = "SELECT fileKey,x,y FROM segyTrace "
    if (None != options.minFileKey) or (None != options.maxFileKey): searchStr += ' WHERE'
    if (None != options.minFileKey): searchStr += ' fileKey>='+str(options.minFileKey)
    if (None != options.minFileKey) and (None != options.maxFileKey): searchStr += ' and'
    if (None != options.maxFileKey): searchStr += ' fileKey<='+str(options.maxFileKey)
    searchStr += ';'

    if options.verbosity >= BOMBASTIC: print searchStr
    
    cu.execute(searchStr)
    lastFileKey = -1
    fileOffset = 0
    for row in cu.fetchall():
        fileKey = row['fileKey']
        x = float(row['x'])
        y = float(row['y'])
        # Write sentinals between each file
        if (fileKey != lastFileKey):
            if -1 != lastFileKey:
                out.write('9999 9999\n')
            lastFileKey = fileKey
            fileOffset = 0
            
        if 0==fileOffset % options.nthShotLabel:
            if options.skipZeros and (int(x) == 0) and (int(y) == 0):
                if options.verbosity>=BOMBASTIC: print 'skipping 0,0 nav point...'
                continue # Don't increment offset... search for first nav point
            lon,lat = ct.getLonLat(x,y)
            out.write (str(lon)+' '+str(lat)+'\n')
            
        fileOffset += 1
    return

########################################

def buildFledermausPoints(cx, options):
    """
    Build a fledermaus x,y point file (no sentinels)

    cx - database connection
    options.fledermaus - file to write out to
    options.nthShotLabel - how often to write out points
    options.key - What to index off of
    """

    out = open(options.fledermausPts,'w')

    ct = CoordTransform(cx,options)
    cu = cx.cursor()  # Database cursor

    if options.verbosity >= TRACE:
        print 'fileKeys: ', options.minFileKey, options.maxFileKey
    
    searchStr  = "SELECT "+options.key+" as key,x,y FROM segyTrace "
    searchStr += "WHERE key % "+str(options.nthShotLabel)+"= 0"
    if (None != options.minFileKey): searchStr += ' fileKey>='+str(options.minFileKey)
    if (None != options.minFileKey) and (None != options.maxFileKey): searchStr += ' and'
    if (None != options.maxFileKey): searchStr += ' fileKey<='+str(options.maxFileKey)
    searchStr += ';'

    if options.verbosity >= BOMBASTIC: print searchStr
    
    cu.execute(searchStr)
    for row in cu.fetchall():
        key = row['key']
        x = float(row['x'])
        y = float(row['y'])
            
        if 0==(key % options.nthShotLabel):
            if options.skipZeros and (x == 0) and (y == 0):
                if options.verbosity>=BOMBASTIC: print 'skipping 0,0 nav point...'
                continue
            lon,lat = ct.getLonLat(x,y)
            out.write (str(lon)+' '+str(lat)+'\n')

    return

########################################

def printGmtSampleScript(options):
    """
    FIX: would be nice to give them more.
    """
    print "# Add the following lines to your gmt script or what was produced by mbm_grdplot"
    print
    print "# Draw the ship track for segy data"
    print "# -M is for multi seg files with '>' between"
    print "psxy "+options.shotLineFilename+" -W1/255/0/255 -M \\"
    print "        -J$MAP_PROJECTION$MAP_SCALE \\"
    print "        -R$MAP_REGION \\"
    print "        -K -O -V >> $PS_FILE"
    print
    print "# put a mark for the shot labels"
    print "psxy "+options.shotTextFilename+"  -Sc.05c   \\"
    print "        -G255/100/0 \\"
    print "        -J$MAP_PROJECTION$MAP_SCALE \\"
    print "        -R$MAP_REGION \\"
    print "        -K -O -V >> $PS_FILE"
    print
    print "# Label every nth shot"
    print "pstext "+options.shotTextFilename+" -G0/0/0 \\"
    print "        -J$MAP_PROJECTION$MAP_SCALE \\"
    print "        -R$MAP_REGION \\"
    print "        -K -O -V >> $PS_FILE"


if __name__ == "__main__":
    (options,args) = myparser.parse_args()

    if options.verbosity >= TRACE:
        print "Verbosity = ", options.verbosity

    if options.fromUTM:
        if None == options.utmZone:
            sys.exit("ERROR: you must specify the UTM zone")
        if (type(10) != type(options.utmZone)):
            sys.exit("FIX: zone must be an int")
        from pyproj import Proj
        params = {}
        params['proj'] = 'utm'
        params['zone'] = options.utmZone
        options.proj = Proj(params)
    
    # FIX: check for database first
    dbName = options.dbfilename
    cx = sqlite.connect(dbName)
    if not os.access(dbName,os.R_OK):
        print "ERROR: database does not exist...",dbName
        sys.exit("No shoes, no shirt, no service")


    if None != options.fledermaus:
        if options.verbosity >= TRACE: print 'Building fledermaus xy file ',options.fledermaus
        buildFledermausLines(cx,options)
    if None != options.fledermausPts:
        if options.verbosity >= TRACE: print 'Building fledermaus xy point file ',options.fledermaus
        buildFledermausPoints(cx,options)

    if (None != options.fledermaus) or (None != options.fledermausPts):
        if options.verbosity >= TRACE: print 'Fledermaus options inhibit other file types'
        sys.exit() # Skip other file types


    buildShotText(cx,options,options.shotTextFilename)

    if not options.skipLine:
        buildShotLines(cx, options, options.shotLineFilename)

    if options.makeGmtScript:
        printGmtSampleScript(options)

